//#define DIAGNOSTICS

using System;
using System.IO.Ports;
using System.Runtime.CompilerServices;
using System.Text;
using System.Threading;
using Microsoft.SPOT.Hardware;
using GHIElectronics.NETMF.Hardware;
using GHIElectronics.NETMF.FEZ;

namespace RobotRemote
{
    /// <summary>
    /// Remote control robot
    /// Receives simple commands from the serial port and acts on them.
    /// Also includes bonus "Cowardly" behaviour.
    /// 
    /// Rob Miles
    /// </summary>
    class RobotDriver
    {
        private FEZ_Pin.AnalogIn DistancePin = FEZ_Pin.AnalogIn.An1;
        private FEZ_Components.DistanceDetector distanceDetector;

        private FEZ_Pin.Digital GreenLedPin = FEZ_Pin.Digital.An4;
        private FEZ_Pin.Digital RedLedPin = FEZ_Pin.Digital.An5;

        private FEZ_Components.LED greenLed;
        private FEZ_Components.LED redLed;

        private FEZ_Pin.Interrupt killSwitchPin = FEZ_Pin.Interrupt.Di4;
        private FEZ_Components.Button killSwitch;

        private FEZ_Pin.Interrupt resetSwitchPin = FEZ_Pin.Interrupt.Di2;
        private FEZ_Components.Button resetSwitch;

        private SerialPort xbeePort;

        private bool systemShutdown = false;

        private string diagnosticMessage = "OK";

#if DIAGNOSTICS

        Timer dumpTimer;

        private void diagnosticTick(object o)
        {
            writeDiagnostics();
        }


        private void setupDiagnostics()
        {
            dumpTimer = new Timer(new TimerCallback(diagnosticTick), this, 500, 500);
        }

        private void writeDiagnostics()
        {
            writeMessage (diagnosticMessage + "\r\n");
        }

#endif
        byte[] statusBytes;

        private void writeMessage(string message)
        {
            statusBytes = Encoding.UTF8.GetBytes(message);
            xbeePort.Write(statusBytes, 0, statusBytes.Length);
        }

        public void SetupRobot()
        {
            distanceDetector =
                new FEZ_Components.DistanceDetector(
                    DistancePin,
                    FEZ_Components.DistanceDetector.SharpSensorType.GP2D120);

            greenLed = new FEZ_Components.LED(GreenLedPin);
            redLed = new FEZ_Components.LED(RedLedPin);
            greenLed.ShutOff();
            redLed.ShutOff();

            killSwitch = new FEZ_Components.Button(killSwitchPin);
            resetSwitch = new FEZ_Components.Button(resetSwitchPin);

            killSwitch.ButtonPressEvent +=
                new FEZ_Components.Button.ButtonPressEventHandler(killSwitch_ButtonPressEvent);

            resetSwitch.ButtonPressEvent +=
                new FEZ_Components.Button.ButtonPressEventHandler(resetSwitch_ButtonPressEvent);

            xbeePort = FEZ_Components.Fez_Xbee.GetPort(true);

#if DIAGNOSTICS
            setupDiagnostics();
#endif

        }

        #region Kill Switch Behaviours

        void shutDown()
        {
            // This is not thread safe. Need to fix for production code
            FEZMini_Robot.Move(0, 0);
            systemShutdown = true;
        }

        void restart()
        {
            systemShutdown = false;
        }

        void resetSwitch_ButtonPressEvent(FEZ_Pin.Interrupt pin, FEZ_Components.Button.ButtonState state)
        {
            if (state == FEZ_Components.Button.ButtonState.Pressed)
            {
                restart();
            }
        }

        void killSwitch_ButtonPressEvent(FEZ_Pin.Interrupt pin, FEZ_Components.Button.ButtonState state)
        {
            if (state == FEZ_Components.Button.ButtonState.Pressed)
            {
                shutDown();
            }
        }

        #endregion

        #region Cowardly Behaviours

        public void CowardlyBehavior()
        {
            while (true)
            {
                float distance = distanceDetector.GetDistance_cm();

                diagnosticMessage = "Distance: " + distance.ToString();

                System.Threading.Thread.Sleep(200);

                if (systemShutdown)
                {
                    greenLed.TurnOn();
                    redLed.TurnOn();
                    continue;
                }

                if (distance > 10)
                    greenLed.TurnOn();
                else
                    greenLed.ShutOff();

                if (distance < 10)
                    redLed.TurnOn();
                else
                    redLed.ShutOff();

                if (distance < 4)
                {
                    FEZMini_Robot.MoveRamp(-100, -100, 1);
                    System.Threading.Thread.Sleep(300);
                    FEZMini_Robot.Move(0, 0);
                }
            }
        }

        #endregion

        #region Remote Control Behaviour

        Decoder decoder = UTF8Encoding.UTF8.GetDecoder();

        public void RemoteControl()
        {
            xbeePort.DataReceived += new SerialDataReceivedEventHandler(xbeePort_DataReceived);
            while (true)
            {
                Thread.Sleep(1000);
            }
        }

        const int BUFFER_SIZE = 100;

        char[] charBuffer = new char[BUFFER_SIZE];

        int bytesUsed;
        int charsUsed;
        bool complete;

        string messageString = "";

        void xbeePort_DataReceived(object sender, SerialDataReceivedEventArgs e)
        {
            if (e.EventType == SerialData.Chars)
            {
                int noAvailable = xbeePort.BytesToRead;
                byte[] byteBuffer = new byte[noAvailable];
                xbeePort.Read(byteBuffer, 0, noAvailable);
                decoder.Convert(byteBuffer, 0, noAvailable,
                    charBuffer, 0, BUFFER_SIZE,
                    false, out bytesUsed, out charsUsed, out complete);
                messageString += new string(charBuffer);
                decodeMessage();
            }
        }

        private void decodeMessage()
        {
            int start = messageString.IndexOf('<', 0);
            if (start >= 0)
            {
                int end = messageString.IndexOf('>', start);
                if (end >= 0)
                {
                    string message = messageString.Substring(start + 1, end - start - 1);

                    switch (message)
                    {
                        case "Forward":
                            redLed.TurnOn();
                            greenLed.TurnOn();
                            FEZMini_Robot.MoveRamp(100, 100, 1);
                            System.Threading.Thread.Sleep(300);
                            FEZMini_Robot.Move(0, 0);
                            redLed.ShutOff();
                            greenLed.ShutOff();
                            break;
                        case "Back":
                            redLed.TurnOn();
                            greenLed.TurnOn();
                            FEZMini_Robot.MoveRamp(-100, -100, 1);
                            System.Threading.Thread.Sleep(300);
                            FEZMini_Robot.Move(0, 0);
                            redLed.ShutOff();
                            greenLed.ShutOff();
                            break;
                        case "Stop":
                            redLed.ShutOff();
                            greenLed.ShutOff();
                            FEZMini_Robot.Move(0, 0);
                            System.Threading.Thread.Sleep(300);
                            FEZMini_Robot.Move(0, 0);
                            break;
                        case "Left":
                            redLed.TurnOn();
                            FEZMini_Robot.MoveRamp(100, -100, 1);
                            System.Threading.Thread.Sleep(300);
                            FEZMini_Robot.Move(0, 0);
                            redLed.ShutOff();
                            break;
                        case "Right":
                            greenLed.TurnOn();
                            FEZMini_Robot.MoveRamp(-100, 100, 1);
                            System.Threading.Thread.Sleep(300);
                            FEZMini_Robot.Move(0, 0);
                            greenLed.ShutOff();
                            break;
                        case "Distance":
                            float distance = distanceDetector.GetDistance_cm();
                            writeMessage("<" + distance.ToString()+">");
                            break;
                    }
                    diagnosticMessage = message;
                    messageString = messageString.Substring(end + 1);
                }
            }
        }

        #endregion
    }
}
